import dataclasses
import os.path
from typing import List

import coko.classes.exceptions as exceptions


@dataclasses.dataclass
class FileOwnership:
    uid: int
    guid: int
    permissions: int
    permissions_octal: bool = dataclasses.field(default=False)

    def __post_init__(self)-> None:
        # We let user to enter permissions in octal but we convert to int
        # because system functions we use deal with them as int.
        #
        # User only would enter permissions in octal in automated tests
        # contexts. In production coko will get native permissions
        # directly from files, so what it would read would be int (from st_mode).
        if self.permissions_octal:
            self.permissions = int(str(self.permissions), 8)

    def __eq__(self, other)-> bool:
        # permissions_octal should not be compared for equality because it is
        # not actually related to file but just a switch to create automated
        # tests. If we would let default __eq__ automated test would fail.
        return all([self.uid == other.uid,
                    self.guid == other.guid,
                    self.permissions == other.permissions])


class Folder(object):
    """A descriptor that sets and returns system folders checking folder
    actually exists.
    """
    def __init__(self):
        self._folder_path: str = None

    def __get__(self, obj, objtype)-> str:
        return self._folder_path

    def __set__(self, obj, value):
        absolute_path: str = os.path.abspath(value)
        if os.path.isdir(absolute_path):
            self._folder_path: str = os.path.abspath(value)
        else:
            raise exceptions.FolderNotFound(absolute_path)


class Configuration:
    source_folder: Folder = Folder()
    destination_folder: Folder = Folder()

    def __init__(self, source_folder: str, destination_folder: str,
                 default_ownership: List):
        self.source_folder = source_folder
        self.destination_folder = destination_folder
        if default_ownership is not None:
            self.default_ownership: FileOwnership = FileOwnership(default_ownership[0],
                                                                  default_ownership[1],
                                                                  default_ownership[2],
                                                                  default_ownership[3])
        else:
            self.default_ownership = None





import sys


def verify_python_version(major_version: int, minor_revision: int)-> None:
    """ Check if current python interpreter version is what we need.
    Compare current interpreter version with the one is provided as needed one.
    If not over needed one version abort execution gracefully warning user
    she is trying to use an unsupported python interpreter.

    :param major_version: Needed major version for Python interpreter.
    :param minor_revision: Needed minor version for Python interpreter.
    :return: None
    """
    if (major_version, minor_revision) > sys.version_info:
        message: str  = f"This program can only be run on Python {major_version}" \
                  f".{minor_version} or newer, you are trying to use Python " \
                  f"{sys.version_info[0]}.{sys.version_info[1]} instead.\nAborting " \
                  "execution."
        sys.exit(message)


import dataclasses
import os
import shutil
from typing import List, Dict

import coko.classes.configuration as configuration


@dataclasses.dataclass
class FileInfo:
    relative_filename_path: str
    ownership: configuration.FileOwnership


def get_files(root_folder: str)-> str:
    """ Iterator to get all files inside a nested folder tree.

    :param root_folder: Base folder where to star getting files from.
    :return: Iterator return files path names relatives to root folder.
    """
    for subdir, dirs, files in os.walk(root_folder):
        for file in files:
            absolute_path = os.path.join(subdir, file)
            relative_path = os.path.relpath(absolute_path, root_folder)
            yield relative_path


def set_ownership(file_path: str, file_info: configuration.FileOwnership)-> None:
    """ Set file for uid, guid and permissions given through a FileOwnership
    object.

    To use this function application should be run as sudo, otherwise the only
    you'll get is an "Operation not permitted" error.

    :param file_path: Absolute file path whose permissions we want to set.
    :param file_info: UID, GUID and access permission to set for this file.
    :return: None
    """
    os.chown(file_path, file_info.uid, file_info.guid)
    os.chmod(file_path, file_info.permissions)


def register_destination_files(config: configuration.Configuration)-> List[FileInfo]:
    """ Walk through destination folder taking note of relative path and
    permissions of every file it meets.

    :param config: Configuration generated from console parameters.
    :return: A list with a FileInfo object for every file in destination folder.
    """
    destination_files = []
    for destination_file in get_files(config.destination_folder):
        file_info = os.stat(os.path.join(config.destination_folder, destination_file))
        permissions = configuration.FileOwnership(file_info.st_uid,
                                                  file_info.st_gid,
                                                  file_info.st_mode)
        destination_files.append(FileInfo(destination_file,
                                          permissions))
    return destination_files


def _copy(source: str, destination: str, permissions: configuration.FileOwnership)-> None:
    """ Copy a file from source to destination and set at destination given
    permissions.

    :param source: Source absolute file path.
    :param destination: Destination absolute file path.
    :param permissions: File permissions to be set at destination.
    :return: None
    """
    shutil.copy2(source, destination)
    set_ownership(destination, permissions)

    # set_ownership() would not be needed here because shutil.copyfile
    # does not overwrite destination file metadata.
    #
    # shutil.copyfile(absolute_source_path, absolute_destination_path)
    #
    # TODO: Refactor code to remove storing destination files metadata.
    # If shutil.copyfile does not overwrite destination files metadata
    # then register_destination_files does not need to save them
    # before copy process. That leads to the point where I need
    # to think if configuration.FileOwnership, set_ownership() and
    # even register_destination_nodes are actually needed any longer.


def copy_files(config: configuration.Configuration,
               destination_ownerships: List[FileInfo])-> None:
    """ Walks through original folder copying into destination but keeping
    original destination folder files permissions.

    It won't copy a file into destination folder if that folder was not already
    present unless Configuration.ownership was other than None. If
    Configuration.ownership wasn't None then ownership would be a
    configuration.FileOwnership object with permissions every newly created
    file at destination will have by default.

    :param config: configuration generated from console parameters.
    :param destination_ownerships: A list with ownership info for every file at
    destination folder.
    :return: None
    """
    destination_files: Dict[str, configuration.FileOwnership] = {item.relative_filename_path: item.ownership
                                                                 for item in destination_ownerships}
    for source_file in get_files(config.source_folder):
        absolute_source_path = os.path.join(config.source_folder, source_file)
        absolute_destination_path = os.path.join(config.destination_folder, source_file)
        # TODO: Add a timestamp check to copy only updated files.
        if source_file in destination_files:
            # Standard use case: copy only files already present at destination.
            _copy(absolute_source_path, absolute_destination_path,
                  destination_files[source_file])
        elif config.default_ownership is not None:
            # Exceptional case: Create files at destination that were present at
            # source folder but not at destination one. We only get here
            # if "--create" flag was used in console command.
            _copy(absolute_source_path, absolute_destination_path,
                  config.default_ownership)







import sys
from typing import List

import coko.classes.console_parser as console_parser
import coko.classes.sync as sync


def main(args: List=sys.argv[1:])-> None:
    config = console_parser.parse_arguments(args)
    destination_ownerships = sync.register_destination_files(config)
    sync.copy_files(config, destination_ownerships)


if __name__ == "__main__":
    main()




import argparse
import sys

import coko.classes.configuration as configuration
import coko.classes.exceptions as exceptions


def parse_arguments(args: list=None) -> configuration.Configuration:
    """ Parse console arguments to generate a Configuration object we
    can work with.

    :param args: Argument list. Usually it's the command you entered but split by spaces.
    :return: Configuration object generated from parsed parameters console command.
    """
    arg_parser = argparse.ArgumentParser(description="A tool to overwrite directories "
                                                     "using files from a different "
                                                     "owners but keeping original "
                                                     "owners and permissions.\n",
                                         epilog="Follow coko development at: "
                                                "<https://github.com/dante-signal31/coko>")
    arg_parser.add_argument(dest="source_folder",
                            metavar="SOURCE_FOLDER",
                            nargs=1, default=None,
                            help="Source folder where you want copy files from.")
    arg_parser.add_argument(dest="destination_folder",
                            metavar="DESTINATION_FOLDER",
                            nargs=1, default=None,
                            help="Destination folder where you want copy "
                                 "files to.")
    arg_parser.add_argument("-c", "--create", dest="default_ownership",
                            nargs=3, type=int,
                            metavar=("UID", "GID", "PERMISSION"),
                            default=None,
                            help="Copy over files not present at destination "
                                 "folder yet and set for them given uid and gid " \
                                 "and permission.")

    # Parse_args returns each parameter in a list. We must take them out so
    # every value at dict is a string.
    parsed_arguments = {item: (value[0] if item != "default_ownership"
                               else value)
                        for (item, value) in vars(arg_parser.parse_args(args)).items()}
    # User enter permission in octal, so I must add a True.
    if  parsed_arguments["default_ownership"] is not None:
        parsed_arguments["default_ownership"].append(True)
    try:
        config = configuration.Configuration(parsed_arguments["source_folder"],
                                             parsed_arguments["destination_folder"],
                                             parsed_arguments["default_ownership"])
    except exceptions.FolderNotFound as e:
        print(f"Folder not found: {e.incorrect_path}")
        sys.exit(1)
    return config



from glob import glob
import os
from setuptools import setup, find_packages  # Always prefer setuptools over distutils
from typing import List

import ci_scripts.ci_tools as tools


VDIST_PACKAGES_CONFIG = "packaging/coko_vdist.cnf"

LONG_DESCRIPTION = """Sometimes you have a directory full of files you want to overwrite periodically.

You may not want to edit those files directly but let other users edit a local copy of those files to copy them over original ones.

Problem there is that local copies may have changed their owners or permissions to be edited so those metadata are carried over original directory overwriting it.

This tools let you take an snapshot of your files metadata in a particular directory in order to restore those metadata after files have been restored.
More info in: https://github.com/dante-signal31/coko
"""


def find_folders_with_this_name(dir_name: str)-> str:
    """ Look for folder with given name, searching from current working dir.

    :param dir_name: Folder name to look for.
    :return: Relative path, from current working dir, where folder is.
    """
    for dir, dirs, files in os.walk('.'):
        if dir_name in dirs:
            yield os.path.relpath(os.path.join(dir, dir_name))


def find_man_pages()-> List:
    """ Look for every folder named "man", inside current working dir, and include
    its files relatives paths in a list suitable to be passed to a data_files
    setup.py parameter. This list will set those manpages to be installed to
    "share/man/man<section>" at target box.

    This code was inspired by a really interesting answer to:
    https://github.com/pypa/packaging-problems/issues/72#issuecomment-279162312

    :return: A list with tuples ("share/man/man<section>", local_relative_path_to_manpage)
    """
    data_files = []
    man_sections = {}
    for dir in find_folders_with_this_name('man'):
        for file in os.listdir(dir):
            section = file.split('.')[-1]
            man_sections[section] = man_sections.get(section, []) + [os.path.join(dir, file)]
    for section in man_sections:
        data_files.append(('share/man/man'+section, man_sections[section]))
    return data_files


def find_info_pages()-> List:
    """ Look for every folder named "info", inside current working dir, and include
    its files relatives paths in a list suitable to be passed to a data_files
    setup.py parameter. This list will set those infopages to be installed to
    "share/info" at target box.

    This code was inspired by a really interesting answer to:
    https://github.com/pypa/packaging-problems/issues/72#issuecomment-279162312

    :return: A list with tuples ("share/info", local_relative_path_to_infopage)
    """
    data_files = []
    info_pages = {}
    for dir in find_folders_with_this_name('info'):
        for file in glob(os.path.join(dir, '*.info')):
            info_pages[dir] = info_pages.get(dir, []) + [file]
    for dir in info_pages:
        data_files.append(('share/info', info_pages[dir]))
    return data_files


setup(name="coko",
      version=tools.get_current_version(VDIST_PACKAGES_CONFIG),
      description="This tools let you take an snapshot of your files metadata "
                  "in a particular directory in order to restore those metadata "
                  "after files have been restored.",
      long_description=LONG_DESCRIPTION,
      author="Dante Signal31",
      author_email="dante.signal31@gmail.com",
      license="BSD-3",
      url="https://github.com/dante-signal31/coko",
      download_url="https://github.com/dante-signal31/coko/releases",
      classifiers=['Development Status :: 4 - Beta',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Information Technology',
                   'Intended Audience :: Science/Research',
                   'Intended Audience :: System Administrators',
                   'Intended Audience :: Telecommunications Industry',
                   'Intended Audience :: Other Audience',
                   'Topic :: System',
                   'Topic :: Security',
                   'License :: OSI Approved :: BSD License',
                   'Operating System :: POSIX :: Linux',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.7'],
      keywords="copy over keeping ownership",
      install_requires=[],
      zip_safe=False,
      packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*",
                                      "tests", "*tests*", "ci_scripts",
                                      "ci_scripts.*", "*.ci_scripts",
                                      "*.ci_scripts.*", "*ci_scripts*",
                                      "ci_scripts*"]),
      data_files=find_man_pages(),
      entry_points={'console_scripts': ['coko=coko.launcher:main', ], }
      )
